/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;

import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConnectionCompositor;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.ConnectionCompositorServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.base.EObjectAware2ServiceBase;

/**
 * This class implements the {@link IConnectionCompositorService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CAC9C66637929C62805C4A11DEE2D9E4
 */
public class ConnectionCompositorService extends
		EObjectAware2ServiceBase<IConnectionCompositor<EObject, EObject, EObject>> implements
		IConnectionCompositorService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final ConnectionCompositorService INSTANCE = new ConnectionCompositorService();

	/** Returns singleton instance of the service. */
	public static ConnectionCompositorService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelConnectionCompositor";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelConnectionCompositor";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "modelConnectionCompositor";

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public void registerConnectionCompositor(
			IConnectionCompositor<EObject, EObject, EObject> compositor, Class<?> sourceClass,
			Class<?> targetClass) {
		addHandler(sourceClass, targetClass, compositor);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the registration of connection compositors." +
				"\nConnection compositors are responsible for linking two model elements," +
				"\nwhile adding the link element as a child to a third model element, e.g.," +
				"\ncreating an edge between two graph nodes and adding it to the set of edges stored with the graph element." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public boolean canConnect(EObject source, EObject target, EObject parent,
			IConnectionCompositionContext context) {
		return findWorkingConnector(source, target, parent, context) != null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean connect(EObject source, EObject target, EObject parent,
			IConnectionCompositionContext context) {
		IConnectionCompositor<EObject, EObject, EObject> connector =
				findWorkingConnector(source, target, parent, context);
		return connector.connect(source, target, parent, context);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canReconnect(EObject connection, EObject source, EObject target,
			IConnectionCompositionContext context) {
		return findWorkingReconnector(connection, source, target, context) != null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean reconnect(EObject connection, EObject source, EObject target,
			IConnectionCompositionContext context) {
		IConnectionCompositor<EObject, EObject, EObject> connector =
				findWorkingReconnector(connection, source, target, context);
		return connector.reconnect(connection, source, target, context);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canDisconnect(EObject connection) {
		IConnectionCompositor<EObject, EObject, EObject> dc = findWorkingDisconnector(connection);
		return dc != null && dc.canDisconnect(connection);
	}

	/** {@inheritDoc} */
	@Override
	public boolean disconnect(EObject connection) {
		IConnectionCompositor<EObject, EObject, EObject> dc = findWorkingDisconnector(connection);
		return dc != null && dc.disconnect(connection);
	}

	/**
	 * Returns the first {@link IConnectionCompositor} which can connect the
	 * given elements (or <code>null</code> if no such connector exists).
	 */
	private IConnectionCompositor<EObject, EObject, EObject> findWorkingConnector(EObject source,
			EObject target, EObject parent, IConnectionCompositionContext context) {
		List<IConnectionCompositor<EObject, EObject, EObject>> list =
				getRegisteredHandlers(source.getClass(), target.getClass());
		if(list == null) {
			return null;
		}
		for(IConnectionCompositor<EObject, EObject, EObject> connector : list) {
			try {
				if(connector.canConnect(source, target, parent, context)) {
					return connector;
				}
			} catch(ClassCastException e) {
				// the container class is not ensured to be compatible. so we ignore all such kinds
				// of exceptions here
			}
		}
		return null;
	}

	/**
	 * Returns the first {@link IConnectionCompositor} which can reconnect the
	 * given elements (or <code>null</code> if no such connector exists).
	 */
	private IConnectionCompositor<EObject, EObject, EObject> findWorkingReconnector(
			EObject connection, EObject source, EObject target,
			IConnectionCompositionContext context) {
		List<IConnectionCompositor<EObject, EObject, EObject>> list =
				getRegisteredHandlers(source.getClass(), target.getClass());
		if(list == null) {
			return null;
		}
		for(IConnectionCompositor<EObject, EObject, EObject> connector : list) {
			if(connector.canReconnect(connection, source, target, context)) {
				return connector;
			}
		}
		return null;
	}

	/**
	 * Returns the first {@link IConnectionCompositor} which can disconnect the
	 * given elements (or <code>null</code> if no such connector exists).
	 */
	private IConnectionCompositor<EObject, EObject, EObject> findWorkingDisconnector(
			EObject connection) {
		for(Class<?> c : handlersBySource.keySet()) {
			Map<Class<?>, List<IConnectionCompositor<EObject, EObject, EObject>>> tmap =
					handlersBySource.get(c);
			for(Class<?> c2 : tmap.keySet()) {
				List<IConnectionCompositor<EObject, EObject, EObject>> list = tmap.get(c2);
				for(IConnectionCompositor<EObject, EObject, EObject> comp : list) {
					if(comp.canDisconnect(connection)) {
						return comp;
					}
				}
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Connection Compositor Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConnectionCompositorServiceIntrospectionDetailsItem(handlersBySource);
	}
}
